/*
 * File:        SIDLObjA.c
 * Package:     SIDL Python Object Adaptor
 * Copyright:   (c) 2001 The Regents of the University of California
 * Release:     $Name: release-0-6-1 $
 * Revision:    @(#) $Revision: 1.18 $
 * Date:        $Date: 2001/11/14 17:32:33 $
 * Description: Python extension type written in C for SIDL object/interface
 *
 * This is a Python extension type written in C to wrap instances of
 * SIDL objects or interfaces in a Python object.  If this looks
 * mysterious to you, look at Programming Python and the BABEL
 * documentation. 
 */

#define SIDLOBJA_MODULE 1
#include "SIDLObjA.h"
#include "SIDL/BaseClass_IOR/BaseClass_IOR.h"
#include "SIDL/BaseException_IOR/BaseException_IOR.h"

#include <string.h>

/**
 * This defines the Python object wrapper for a SIDL object or interface.
 * This one Python extension type written in C can support an arbitrary
 * class or interface. The peculiarities of a particular class or interface
 * are stored in the d_methods member which defines what the object
 * can do.
 */
struct SIDLPythonObject {
  PyObject_HEAD                 /* standard Python object header */

  /* SIDL specific extensions */

  PyMethodDef                         *d_methods;
  struct SIDL_BaseClass__object *    (*d_getIOR)(struct SIDLPythonObject *);
  union {
    struct SIDL_BaseClass__object     *d_ior;
    struct SIDL_BaseInterface__object *d_abint;
  }                                    d_sidlobj;
  int				       d_refType;
};

/**
 * This <code>typedef</code> is required by PyObject_NEW; otherwise,
 * it is unused.
 */
typedef struct SIDLPythonObject SPObject;

/**
 * A function to get the independent object representation (IOR) for
 * a wrapped instance of a class.
 */
static struct SIDL_BaseClass__object *
SIDLPython_Object_getIOR(struct SIDLPythonObject *sobj) {
  return sobj->d_sidlobj.d_ior;
}

/**
 * A function to get the independent object representation (IOR) for
 * a wrapped instance of a interface.
 */
static struct SIDL_BaseClass__object *
SIDLPython_Interface_getIOR(struct SIDLPythonObject *sobj) {
  return sobj->d_sidlobj.d_abint->d_object;
}

/**
 * Provide a forward declaration of the Python type object used
 * by <code>SIDLPythonObject</code>'s.
 */
staticforward PyTypeObject SIDLPythonObjectType;

#define is_SIDLobject(v) ((v)->ob_type == &SIDLPythonObjectType)

/*
 * Exported C API methods
 */

static SIDL_Object_Create_RETURN
SIDL_Object_Create SIDL_Object_Create_PROTO {
  PyObject *result = NULL;
  if (methods) {
    if (ior) {
      struct SIDLPythonObject *self = 
	PyObject_NEW(SPObject, &SIDLPythonObjectType);
      if (self){
	self->d_sidlobj.d_ior = ior;
        if (refType == SIDL_PyNewRef) {
          struct SIDL_BaseClass__object *baseIOR = (*ior->d_epv->f__cast_ior)(ior, "SIDL.BaseClass");
          (*baseIOR->d_epv->f_addReference)(baseIOR);
        }
	self->d_methods = methods;
	self->d_getIOR = SIDLPython_Object_getIOR;
        self->d_refType = refType;
	result = (PyObject *)self;
      }
      else {
        if (refType == SIDL_PyStealRef) {
          struct SIDL_BaseClass__object *baseIOR = (*ior->d_epv->f__cast_ior)(ior, "SIDL.BaseClass");
          (*baseIOR->d_epv->f_deleteReference)(baseIOR);
        }
      }
    }
    else {
      result = Py_None;
      Py_INCREF(result);
    }
  }
  else {
    PyErr_SetString(PyExc_AssertionError,
		    "SIDL object has NULL IOR or vtable");
  }
  return result;
}

static SIDL_Get_IOR_RETURN
SIDL_Get_IOR SIDL_Get_IOR_PROTO {
  if (is_SIDLobject(obj)) {
    return (*((struct SIDLPythonObject *) obj)->d_getIOR)((struct SIDLPythonObject *) obj);
  }
  return NULL;
}

static SIDL_Cast_RETURN
SIDL_Cast SIDL_Cast_PROTO {
  struct SIDL_BaseClass__object *ior = SIDL_Get_IOR(obj);
  return (ior) ? ((*(ior->d_epv->f__cast_ior))(ior, name)) : NULL;
}

static SIDL_Opaque_Create_RETURN
SIDL_Opaque_Create SIDL_Opaque_Create_PROTO {
  return PyCObject_FromVoidPtr(opaque_ptr, NULL);
}

static SIDL_Opaque_Convert_RETURN
SIDL_Opaque_Convert SIDL_Opaque_Convert_PROTO {
  if (PyCObject_Check(obj)) {
    *opaque_ptr = PyCObject_AsVoidPtr(obj);
    return 1;
  }
  return 0;
}

static SIDL_PyType_RETURN
SIDL_PyType SIDL_PyType_PROTO {
  return (PyObject *) &SIDLPythonObjectType;
}

static SIDL_Interface_Create_RETURN
SIDL_Interface_Create SIDL_Interface_Create_PROTO {
  PyObject *result = NULL;
  if (methods) {
    if (abint) {
      struct SIDLPythonObject *self = 
	PyObject_NEW(SPObject, &SIDLPythonObjectType);
      if (self){
	self->d_sidlobj.d_abint = abint;
        if (refType == SIDL_PyNewRef) {
          (*abint->d_epv->f_addReference)(abint->d_object);
        }
	self->d_methods = methods;
	self->d_getIOR = SIDLPython_Interface_getIOR;
        self->d_refType = refType;
	result = (PyObject *)self;
      }
      else {
        if (refType == SIDL_PyStealRef) {
          (*abint->d_epv->f_deleteReference)(abint->d_object);
        }
      }
    }
    else {
      result = Py_None;
      Py_INCREF(result);
    }
  }
  else {
    PyErr_SetString(PyExc_AssertionError,
		    "SIDL interface has NULL vtable");
  }
  return result;
}

static SIDL_Interface_Check_RETURN
SIDL_Interface_Check SIDL_Interface_Check_PROTO {
  if (is_SIDLobject(obj) && 
      (((struct SIDLPythonObject *) obj)->d_getIOR == SIDLPython_Interface_getIOR)) {
    return ((struct SIDLPythonObject *) obj)->d_sidlobj.d_abint;
  }
  return NULL;
}

static SIDL_PyExceptionCast_RETURN
SIDL_PyExceptionCast SIDL_PyExceptionCast_PROTO {
  return ex->d_epv->f__cast_ior(ex, name);
}

/*
 * BASIC OBJECT METHODS
 */

static PyObject *
spoa_getattr(struct SIDLPythonObject *self, char *name)
{
  if (!strcmp(name, "__members__")){
    return Py_BuildValue((char *) "[]");
  }
  return Py_FindMethod(self->d_methods, (PyObject *) self, name);
}

static void
spoa_self_destruct(struct SIDLPythonObject *self)
{
  /* remove Python's reference to this SIDL object */
  if (self->d_refType != SIDL_PyWeakRef) {
    struct SIDL_BaseClass__object *baseIOR = SIDL_Cast((PyObject *) self, "SIDL.BaseClass");
    (*(baseIOR->d_epv->f_deleteReference))(baseIOR);
  }
  self->d_sidlobj.d_ior = NULL;
  self->d_sidlobj.d_abint = NULL;
  self->d_getIOR = NULL;
  self->d_methods = NULL;
  self->d_refType = SIDL_PyWeakRef;
  PyMem_DEL(self);
}

static long
spoa_hash(struct SIDLPythonObject *self)
{
  struct SIDL_BaseClass__object *baseIOR = SIDL_Cast((PyObject *) self, "SIDL.BaseClass");
  /* Return value setup */
  long _result = 0;
  PyObject *_hash_arg = NULL;
  /* Sanity check */
  if (baseIOR == 0) {
    PyErr_SetString(PyExc_TypeError, "Null IOR pointer in spoa_hash()");
  }
  /* Hash IOR pointer */
  _hash_arg = Py_BuildValue((char *) "i", baseIOR->d_object);
  if (!_hash_arg) {
    if (PyErr_Occurred()) PyErr_Print();
    return _result;
  }
  _result = PyObject_Hash(_hash_arg);
  Py_DECREF(_hash_arg);
  /* Return value cleanup */
  return _result;
}

static int spoa__convert(PyObject *obj, struct SIDL_BaseInterface__object **sidlobj) {
  *sidlobj = SIDL_Cast(obj, "SIDL.BaseInterface");
  if (!(*sidlobj) && (obj != Py_None)) {
    PyErr_SetString(PyExc_TypeError, "Argument is not of type SIDL.BaseInterface");
    return 0;
  }
  return 1;
}

static int
spoa_cmp(struct SIDLPythonObject *self, PyObject *arg)
{
  struct SIDL_BaseClass__object *baseIOR = SIDL_Cast((PyObject *) self, "SIDL.BaseClass");
  /* Return value setup */
  PyObject *o1 = NULL;
  PyObject *o2 = NULL;
  int _result = 1;
  /* Parse arguments */
  struct SIDL_BaseInterface__object * iobj = NULL;
  const int _okay = spoa__convert(arg, &iobj);
  if (_okay == 0) {
    PyObject_Print(arg, stdout, 0);
    PyErr_SetString(PyExc_TypeError, "Could not parse arguments for spoa_cmp");
    return _result;
  }
  /* Sanity check */
  if (baseIOR == 0) {
    PyErr_SetString(PyExc_TypeError, "Null IOR pointer in spoa_cmp()");
  }
  /* Dispatch to IOR */
  o1 = Py_BuildValue((char *) "i", baseIOR->d_object);
  if (!o1) {
    if (PyErr_Occurred()) PyErr_Print();
    return _result;
  }
  o2 = Py_BuildValue((char *) "i", iobj->d_object);
  if (!o2) {
    if (PyErr_Occurred()) PyErr_Print();
    return _result;
  }
  _result = PyObject_Compare(o1, o2);
  Py_DECREF(o1);
  Py_DECREF(o2);
  return _result;
}

#if 0
static PyObject *
spoa_str(struct SIDLPythonObject *self)
{
  struct SIDL_BaseClass__object *baseIOR = SIDL_Cast((PyObject *) self, "SIDL.BaseClass");
  /* Return value setup */
  PyObject *_return_value = NULL;
  char *_result = 0;
  /* Dispatch to the IOR */
  _result = (*baseIOR->d_epv->d___str__)();
  /* Get the implementation as a string */
  _return_value = PyString_FromString((const char *) _result);
  /* Return value cleanup */
  Py_INCREF(_return_value);
  return _return_value;
}
#endif

static PyObject *
spoa_richcmp(struct SIDLPythonObject *self, PyObject *arg, int op)
{
  struct SIDL_BaseClass__object *baseIOR = SIDL_Cast((PyObject *) self, "SIDL.BaseClass");
  /* Return value setup */
  PyObject *_return_value = NULL;
  PyObject *o1 = NULL;
  PyObject *o2 = NULL;
  /* Parse arguments */
  struct SIDL_BaseInterface__object * iobj = NULL;
  const int _okay = spoa__convert(arg, &iobj);
  if (_okay == 0) {
    switch(op) {
    case Py_EQ:
      _return_value = Py_False;
      Py_INCREF(_return_value);
      break;
    case Py_NE:
      _return_value = Py_True;
      Py_INCREF(_return_value);
      break;
    case Py_LE:
    case Py_GE:
    case Py_LT:
    case Py_GT:
      PyErr_SetString(PyExc_TypeError, "Could not parse arguments for spoa_richcmp");
    }
    return _return_value;
  }
  /* Sanity check */
  if (baseIOR == 0) {
    PyErr_SetString(PyExc_TypeError, "Null IOR pointer in spoa_richcmp()");
  }
  /* Dispatch to IOR */
  o1 = Py_BuildValue((char *) "i", baseIOR->d_object);
  if (!o1) {
    if (PyErr_Occurred()) PyErr_Print();
    return _return_value;
  }
  o2 = Py_BuildValue((char *) "i", iobj->d_object);
  if (!o2) {
    if (PyErr_Occurred()) PyErr_Print();
    return _return_value;
  }
  _return_value = PyObject_RichCompare(o1, o2, op);
  Py_DECREF(o1);
  Py_DECREF(o2);
  Py_INCREF(_return_value);
  return _return_value;
}

static PyTypeObject SIDLPythonObjectType = {
  /* type header */
  PyObject_HEAD_INIT(NULL)
  0,
  (char *) "SIDLPythonObject",
  sizeof(struct SIDLPythonObject),
  0,

  /* standard methods */
  (destructor) spoa_self_destruct, /* call when reference count == 0 */
  (printfunc) 0,
  (getattrfunc) spoa_getattr,
  (setattrfunc) 0,
  (cmpfunc) spoa_cmp,
  (reprfunc) 0,
 
  /* type categories */
  (PyNumberMethods *) 0,
  (PySequenceMethods *) 0,
  (PyMappingMethods *) 0,

  /* more methods */
  (hashfunc) spoa_hash,
  (ternaryfunc) 0,
  (reprfunc) 0 /*spoa_str*/,
  (getattrofunc) 0,
  (setattrofunc) 0,
  
  (PyBufferProcs *) 0,
  0,
  /* documentation string */
  (char *) "This is a Python wrapper for a SIDL object or interface.",

  (traverseproc) 0,
  (inquiry) 0,
  (richcmpfunc) spoa_richcmp
};

static struct PyMethodDef spoa_methods[] = {
  /* this module exports no methods */
  { NULL, NULL }
};

#ifdef __cplusplus
extern "C" void initSIDLObjA(void);
#else
extern void initSIDLObjA(void);
#endif

void
initSIDLObjA(void)
{
  PyObject *module, *dict, *c_api;
  static void *spoa_api[SIDL_API_pointers];

  SIDLPythonObjectType.ob_type = &PyType_Type;

  module = Py_InitModule((char *) "SIDLObjA", spoa_methods);
  dict = PyModule_GetDict(module);
  spoa_api[SIDL_Object_Create_NUM] = (void *)SIDL_Object_Create;
  spoa_api[SIDL_Get_IOR_NUM] = (void *)SIDL_Get_IOR;
  spoa_api[SIDL_Cast_NUM] = (void *)SIDL_Cast;
  spoa_api[SIDL_Interface_Create_NUM] = (void *)SIDL_Interface_Create;
  spoa_api[SIDL_Interface_Check_NUM] = (void *)SIDL_Interface_Check;
  spoa_api[SIDL_Opaque_Create_NUM] = (void *)SIDL_Opaque_Create;
  spoa_api[SIDL_Opaque_Convert_NUM] = (void *)SIDL_Opaque_Convert;
  spoa_api[SIDL_PyExceptionCast_NUM] = (void *)SIDL_PyExceptionCast;
  spoa_api[SIDL_PyType_NUM] = (void *)SIDL_PyType;
  c_api = PyCObject_FromVoidPtr((void *)spoa_api, NULL);
  if (c_api) {
    PyDict_SetItemString(dict, (char *) "_C_API", c_api);
    Py_DECREF(c_api);
  }
  if (PyErr_Occurred()) {
    Py_FatalError((char *) "Can't initialize module SIDLObjA.");
  }
}
