/*
 * File:        SIDLPyArrays.c
 * Copyright:   (c) 2001 The Regents of the University of California
 * Release:     $Name: release-0-6-1 $
 * Revision:    @(#) $Revision: 1.10 $
 * Date:        $Date: 2001/10/09 15:36:22 $
 * Description: Runtime support routines to convert SIDL arrays to/from Python
 *
 * This file provides functions to convert SIDL arrays to Python with or
 * without borrowing data and functions to convert Python arrays to SIDL
 * with or without borrowing. When borrowing data is not possible, data
 * is copied.
 */

#define SIDLPyArrays_MODULE 1
#include "SIDLPyArrays.h"

#include "SIDL/bool_IOR/bool_IOR.h"
#include "SIDL/char_IOR/char_IOR.h"
#include "SIDL/dcomplex_IOR/dcomplex_IOR.h"
#include "SIDL/double_IOR/double_IOR.h"
#include "SIDL/fcomplex_IOR/fcomplex_IOR.h"
#include "SIDL/float_IOR/float_IOR.h"
#include "SIDL/int_IOR/int_IOR.h"
#include "SIDL/long_IOR/long_IOR.h"
#include "SIDL/opaque_IOR/opaque_IOR.h"
#include "SIDL/string_IOR/string_IOR.h"

/* include Numerical Python header file */
#include "Numeric/arrayobject.h"
#include <stdlib.h>
#include <string.h>

int CopyOpaquePointer(void *, const int *, PyObject *);
int CopyStringPointer(void *, const int *, PyObject *);

/**
 * Convert the array description information from the Numerical Python array
 * into a form that SIDL likes.  SIDL needs a vector of lower and upper
 * index bounds, and a vector holding the stride.
 */
SIDL_array__extract_python_info_RETURN
SIDL_array__extract_python_info SIDL_array__extract_python_info_PROTO
{
  *lower = *upper = *stride = NULL;
  if (PyArray_Check(pya)) {
    const PyArrayObject *array = (PyArrayObject *)pya;
    if (array->nd >= 0) {
      /* a zero dimensional Python array is like a scalar */
      const int dimen = array->nd ? array->nd : 1;
      int i;
      *dimension = dimen;
      *lower = malloc(sizeof(int) * dimen);
      *upper = malloc(sizeof(int) * dimen);
      *stride = malloc(sizeof(int) * dimen);
      **lower = **upper = 0;
      **stride = 1;
      for(i = 0; i < array->nd; ++i) {
        (*lower)[i] = 0;
        (*upper)[i] = array->dimensions[i] - 1;
        (*stride)[i] = array->strides[i]/array->descr->elsize;
        if ((array->strides[i] % array->descr->elsize) != 0) {
          goto error;
        }
      }
      return 1;
    }
  }
 error:
  if (*lower)  free(*lower);
  if (*upper)  free(*upper);
  if (*stride) free(*stride);
  *lower = *upper = *stride = NULL;
  return 0;
}

/**
 * Return a true value iff the array is contiguous and in column-major order
 * (i.e. like FORTRAN).
 *
 *   stride    the stride for each dimension in bytes.
 *   numelem   the number of elements in each dimension.
 *   dim       the number of dimensions.
 *   elsize    the element size in bytes.
 */
static int
columnMajorOrder(const int stride[],
                 const int numelem[],
                 const int dim,
                 const int elsize)
{
  int dimStride = elsize, i;
  for(i = 0; i < dim; ++i) {
    if (stride[i] != dimStride) return FALSE;
    dimStride *= numelem[i];
  }
  return TRUE;
}

/**
 * Return a true value iff the array is contiguous and in row-major order
 * (i.e. like C). 
 *
 *   stride    the stride for each dimension in bytes.
 *   numelem   the number of elements in each dimension.
 *   dim       the number of dimensions.
 *   elsize    the element size in bytes.
 */
static int
rowMajorOrder(const int stride[],
              const int numelem[],
              const int dim,
              const int elsize)
{
  int dimStride = elsize, i;
  for(i = dim - 1; i >= 0; --i) {
    if (stride[i] != dimStride) return FALSE;
    dimStride *= numelem[i];
  }
  return TRUE;
}

/**
 * Return the number elements in the array times elsize.  If elsize is the
 * element size in bytes, this will return the size of the array in bytes.
 * If elsize is 1, this returns the number of elements in the array.
 *
 *   numelem     the number of elements in each dimension. Array of dim
 *               integers.
 *   dim         the number of dimensions.
 *   elsize      to get the size of the array in bytes, pass in the
 *               size of an element in bytes. To get the number of
 *               elements, pass in 1.
 */
static size_t
arraySize(const int numelem[],
          const int dim,
          const int elsize)
{
  size_t result = 0;
  if (dim > 0) {
    int i;
    result = elsize;
    for(i = 0; i < dim; ++i) {
      result *= numelem[i];
    }
  }
  return result;
}

/**
 * A parameterized function to copy one strided array into another.
 */
#define copyData_impl(name, dtype, stype) \
static void copy_ ## name ## _data(char * restrict dest, \
                                   const int dstride[], \
                                   const char * restrict src, \
                                   const int sstride[], \
                                   const int numelem[], \
                                   const int dim) \
{ \
  size_t num = arraySize(numelem, dim, sizeof(stype)); \
  if (num > 0) { \
    if ((sizeof(stype) == sizeof(dtype)) && \
        ((columnMajorOrder(sstride, numelem, dim, sizeof(stype)) && \
          columnMajorOrder(dstride, numelem, dim, sizeof(dtype))) || \
         (rowMajorOrder(sstride, numelem, dim, sizeof(stype)) && \
          rowMajorOrder(dstride, numelem, dim, sizeof(dtype))))) { \
      memcpy(dest, src, num); \
    } \
    else { \
      int * restrict ind = malloc(sizeof(int)*dim), i; \
      memset(ind, 0, sizeof(int)*dim); \
      num /= sizeof(stype); \
      while (num--) { \
        *((dtype * restrict)dest) = *((const stype * restrict)src); \
        for (i = 0; i < dim ; ++i) { \
          dest += dstride[i]; \
          src += sstride[i]; \
          if (++(ind[i]) < numelem[i]) { \
            break; \
          } \
          else { \
            ind[i] = 0; \
            dest -= (numelem[i]*dstride[i]); \
            src -= (numelem[i]*sstride[i]); \
          } \
        } \
      } \
      free(ind); \
    } \
  } \
}

copyData_impl(bool, int, int)
copyData_impl(char, char, char)
copyData_impl(int, int, int)
copyData_impl(tolong, long, int64_t)
copyData_impl(fromlong, int64_t, long)
copyData_impl(float, float, float)
copyData_impl(double, double, double)
copyData_impl(fcomplex, struct SIDL_fcomplex, struct SIDL_fcomplex)
copyData_impl(dcomplex, struct SIDL_dcomplex, struct SIDL_dcomplex)

/*
 * Parameterized function implementation to borrow a particular type
 * array.
 */
#define SIDL_borrow_impl(stype,dtype,pytype) \
{\
  if (PyArray_Check(obj)) {\
    PyArrayObject *pya = (PyArrayObject *)obj;\
    if (pytype == pya->descr->type_num) {\
      int dimension, *lower, *upper, *stride;\
      if (SIDL_array__extract_python_info((PyObject *)pya, &dimension, \
                                          &lower, &upper, &stride)) {\
        *array = SIDL_ ## stype ## __array_borrow((dtype *)pya->data,\
                                         dimension, lower, upper, stride);\
        free(lower);\
        free(upper);\
        free(stride);\
        return (*array != NULL);\
      }\
    }\
  }\
  return SIDL_ ## stype ## __clone_python_array(obj, array);\
}


SIDL_bool__borrow_python_array_RETURN
SIDL_bool__borrow_python_array SIDL_bool__borrow_python_array_PROTO
SIDL_borrow_impl(bool, SIDL_bool, PyArray_INT)

SIDL_char__borrow_python_array_RETURN
SIDL_char__borrow_python_array SIDL_char__borrow_python_array_PROTO
SIDL_borrow_impl(char, char, PyArray_CHAR)

SIDL_dcomplex__borrow_python_array_RETURN
SIDL_dcomplex__borrow_python_array SIDL_dcomplex__borrow_python_array_PROTO
SIDL_borrow_impl(dcomplex, struct SIDL_dcomplex, PyArray_CDOUBLE)

SIDL_double__borrow_python_array_RETURN
SIDL_double__borrow_python_array SIDL_double__borrow_python_array_PROTO
SIDL_borrow_impl(double, double, PyArray_DOUBLE)

SIDL_fcomplex__borrow_python_array_RETURN
SIDL_fcomplex__borrow_python_array SIDL_fcomplex__borrow_python_array_PROTO
SIDL_borrow_impl(fcomplex, struct SIDL_fcomplex, PyArray_CFLOAT)

SIDL_float__borrow_python_array_RETURN
SIDL_float__borrow_python_array SIDL_float__borrow_python_array_PROTO
SIDL_borrow_impl(float, float, PyArray_FLOAT)

SIDL_int__borrow_python_array_RETURN
SIDL_int__borrow_python_array SIDL_int__borrow_python_array_PROTO
SIDL_borrow_impl(int, int, PyArray_INT)

SIDL_long__borrow_python_array_RETURN
SIDL_long__borrow_python_array SIDL_long__borrow_python_array_PROTO
{
  return SIDL_long__clone_python_array(obj,array);
}

SIDL_string__borrow_python_array_RETURN
SIDL_string__borrow_python_array SIDL_string__borrow_python_array_PROTO
{
  return SIDL_string__clone_python_array(obj,array);
}

SIDL_opaque__borrow_python_array_RETURN
SIDL_opaque__borrow_python_array SIDL_opaque__borrow_python_array_PROTO
{
  return SIDL_opaque__clone_python_array(obj,array);
}

/*
 * Parameterized implementation of a clone function.
 */
#define ClonePython_impl(sidlname, copyname, sidltype, pyarraytype) \
{ \
  int result = 0; \
  PyArrayObject *pya = \
    (PyArrayObject *)PyArray_FromObject(obj, pyarraytype, 0, 0); \
  *array = 0; \
  if (pya) { \
    int dimen, *lower, *upper, *stride; \
    if (SIDL_array__extract_python_info((PyObject *)pya, &dimen, \
                                        &lower, &upper, &stride)) { \
      *array = \
         SIDL_ ## sidlname ## __array_create(dimen, lower, upper); \
      if (*array) { \
        int *bytestride = malloc(sizeof(int)*dimen), i; \
        int *numelem = malloc(sizeof(int)*dimen); \
        for(i = 0; i < dimen; ++i){ \
          numelem[i] = 1 + upper[i] - lower[i]; \
          bytestride[i] = sizeof(sidltype)* \
             (*array)->d_stride[i]; \
        } \
        copy_ ## copyname ## _data((char *)((*array)->d_firstElement),  \
                       bytestride, \
                       pya->data, \
                       pya->strides, \
                       numelem, \
                       dimen); \
        free(numelem); \
        free(bytestride); \
        result = 1; \
      } \
      free(stride); \
      free(upper); \
      free(lower); \
    } \
    Py_DECREF(pya); \
  } \
  return result; \
}

SIDL_bool__clone_python_array_RETURN
SIDL_bool__clone_python_array SIDL_bool__clone_python_array_PROTO
ClonePython_impl(bool, bool, SIDL_bool, PyArray_INT)

SIDL_char__clone_python_array_RETURN
SIDL_char__clone_python_array SIDL_char__clone_python_array_PROTO
ClonePython_impl(char, char, char, PyArray_CHAR)

SIDL_dcomplex__clone_python_array_RETURN
SIDL_dcomplex__clone_python_array SIDL_dcomplex__clone_python_array_PROTO
ClonePython_impl(dcomplex, dcomplex, struct SIDL_dcomplex, PyArray_CDOUBLE)

SIDL_double__clone_python_array_RETURN
SIDL_double__clone_python_array SIDL_double__clone_python_array_PROTO
ClonePython_impl(double, double, double, PyArray_DOUBLE)

SIDL_fcomplex__clone_python_array_RETURN
SIDL_fcomplex__clone_python_array SIDL_fcomplex__clone_python_array_PROTO
ClonePython_impl(fcomplex, fcomplex, struct SIDL_fcomplex, PyArray_CFLOAT)

SIDL_float__clone_python_array_RETURN
SIDL_float__clone_python_array SIDL_float__clone_python_array_PROTO
ClonePython_impl(float, float, float, PyArray_FLOAT)

SIDL_int__clone_python_array_RETURN
SIDL_int__clone_python_array SIDL_int__clone_python_array_PROTO
ClonePython_impl(int, int, int, PyArray_INT)

SIDL_long__clone_python_array_RETURN
SIDL_long__clone_python_array SIDL_long__clone_python_array_PROTO
ClonePython_impl(long, fromlong, int64_t, PyArray_LONG)


SIDL_array__convert_python_RETURN
SIDL_array__convert_python SIDL_array__convert_python_PROTO
{
  int result = FALSE;
  if (PyArray_Check(pya_src)) {
    int i, j;
    const int size = PyArray_Size(pya_src);
    const PyArrayObject *pya = (PyArrayObject *)pya_src;
    int *ind = malloc(sizeof(int) * dimen);
    char *pydata = pya->data;
    memset(ind, 0, sizeof(int) * dimen);
    if (size == 1) {              /* handle zero dimensional PyArray */
      if (!((*setfunc)(sidl_dest, ind, *(PyObject **)pydata))) {
        result = TRUE;
      }
    }
    else {
      for(i = 0; i < size; ++i){
        if ((*setfunc)(sidl_dest, ind, *(PyObject **)pydata)) {
          goto error;
        }
        for(j = 0; j < dimen; ++j) {
          pydata += pya->strides[j];
          if (++(ind[j]) < pya->dimensions[j]) {
            break;
          }
          else {
            ind[j] = 0;
            pydata -= (pya->dimensions[j]*pya->strides[j]);
          }
        }
      }
      result = TRUE;
    }
  error:
    free(ind);
  }
  return result;
}

int CopyOpaquePointer(void *array,
                      const int *ind,
                      PyObject *data)
{
  if (PyCObject_Check(data)) {
    SIDL_opaque__array_set((struct SIDL_opaque__array*)array,
                           ind,
                           PyCObject_AsVoidPtr(data));
    return FALSE;
  }
  return TRUE;
}

int CopyStringPointer(void *array,
                      const int *ind,
                      PyObject *data)
{
  PyObject *str = PyObject_Str(data);
  if (str) {
    SIDL_string__array_set((struct SIDL_string__array*)array, ind,
                           PyString_AsString(str));
    Py_DECREF(str);
    return FALSE;
  }
  return TRUE;
}

SIDL_opaque__clone_python_array_RETURN
SIDL_opaque__clone_python_array SIDL_opaque__clone_python_array_PROTO
{
  int result = 0;
  PyObject *pya = PyArray_FromObject(obj, PyArray_OBJECT, 0, 0);
  *array = NULL;
  if (pya) {
    if (PyArray_OBJECT == ((PyArrayObject *)pya)->descr->type_num) {
      int dimen, *lower, *upper, *stride;
      if (SIDL_array__extract_python_info(pya, &dimen, &lower, &upper,
                                          &stride)) {
        *array = SIDL_opaque__array_create(dimen, lower, upper);
        free(stride);
        free(upper);
        free(lower);
        result = SIDL_array__convert_python(pya, dimen, *array,
                                            CopyOpaquePointer);
        if (*array && !result) {
          SIDL_opaque__array_destroy(*array);
          *array = NULL;
        }
      }
    }
    Py_DECREF(pya);
  }
  return result;
}

SIDL_string__clone_python_array_RETURN
SIDL_string__clone_python_array SIDL_string__clone_python_array_PROTO
{
  int result = 0;
  PyObject *pya = PyArray_FromObject(obj, PyArray_OBJECT, 0, 0);
  *array = NULL;
  if (pya) {
    if (PyArray_OBJECT == ((PyArrayObject *)pya)->descr->type_num) {
      int dimen, *lower, *upper, *stride;
      if (SIDL_array__extract_python_info(pya, &dimen, &lower, &upper,
                                          &stride)) {
        *array = SIDL_string__array_create(dimen, lower, upper);
        free(stride);
        free(upper);
        free(lower);
        result = SIDL_array__convert_python(pya, dimen, *array, 
                                            CopyStringPointer);
        if (*array && !result) {
          SIDL_string__array_destroy(*array);
          *array = NULL;
        }
      }
    }
    Py_DECREF(pya);
  }
  return result;
}


#define CloneSIDL_impl(copyname, sidltype, pyarraytype) \
{ \
  PyArrayObject *pya = NULL; \
  if (array) { \
    const int dimen = array->d_dimen; \
    int i; \
    int *numelem = malloc(sizeof(int)*dimen); \
    int *bytestride = malloc(sizeof(int)*dimen); \
    for(i = 0; i < dimen; ++i){ \
      numelem[i] = 1 + array->d_upper[i] - array->d_lower[i]; \
      bytestride[i] = sizeof(sidltype)* \
        array->d_stride[i]; \
    } \
    pya = (PyArrayObject *)PyArray_FromDims(dimen, numelem, pyarraytype); \
    if (pya) { \
      copy_ ## copyname ## _data(pya->data, \
                     pya->strides, \
                     (char *)((array)->d_firstElement),  \
                     bytestride, \
                     numelem, \
                     dimen); \
    } \
    free(numelem); \
    free(bytestride); \
  } \
  return pya ? PyArray_Return(pya) : NULL; \
}

SIDL_bool__python_clone_array_RETURN
SIDL_bool__python_clone_array SIDL_bool__python_clone_array_PROTO
CloneSIDL_impl(bool, SIDL_bool, PyArray_INT)

SIDL_char__python_clone_array_RETURN
SIDL_char__python_clone_array SIDL_char__python_clone_array_PROTO
CloneSIDL_impl(char, char, PyArray_CHAR)

SIDL_dcomplex__python_clone_array_RETURN
SIDL_dcomplex__python_clone_array SIDL_dcomplex__python_clone_array_PROTO
CloneSIDL_impl(dcomplex, struct SIDL_dcomplex, PyArray_CDOUBLE)

SIDL_double__python_clone_array_RETURN
SIDL_double__python_clone_array SIDL_double__python_clone_array_PROTO
CloneSIDL_impl(double, double, PyArray_DOUBLE)

SIDL_fcomplex__python_clone_array_RETURN
SIDL_fcomplex__python_clone_array SIDL_fcomplex__python_clone_array_PROTO
CloneSIDL_impl(fcomplex, struct SIDL_fcomplex, PyArray_CFLOAT)

SIDL_float__python_clone_array_RETURN
SIDL_float__python_clone_array SIDL_float__python_clone_array_PROTO
CloneSIDL_impl(float, float, PyArray_FLOAT)

SIDL_int__python_clone_array_RETURN
SIDL_int__python_clone_array SIDL_int__python_clone_array_PROTO
CloneSIDL_impl(int, int, PyArray_INT)

SIDL_long__python_clone_array_RETURN
SIDL_long__python_clone_array SIDL_long__python_clone_array_PROTO
CloneSIDL_impl(tolong, int64_t, PyArray_LONG)

SIDL_array__convert_sidl_RETURN
SIDL_array__convert_sidl SIDL_array__convert_sidl_PROTO
{
  if (PyArray_Check(pya_dest)) {
    PyArrayObject *pya = (PyArrayObject *)pya_dest;
    size_t size = arraySize(numelem, dimen, 1);
    int i;
    char *dest = pya->data;
    while (size--) {
      if ((*getfunc)(sidl_src, lower, (PyObject **)dest))
        return FALSE;
      for(i = 0; i < dimen; ++i) {
        dest += pya->strides[i];
        if (++(lower[i]) <= upper[i]) {
          break;
        }
        else {
          dest -= pya->strides[i]*numelem[i];
          lower[i] -= numelem[i];
        }
      }
    }
    return TRUE;
  }
  return FALSE;
}

static int
getAndConvertString(void *array, const int *ind, PyObject **dest)
{
  char *str = 
    SIDL_string__array_get((struct SIDL_string__array *)array, ind);
  if (str) {
    *dest = PyString_FromString(str); 
    free(str);
  }
  else {
    Py_INCREF(Py_None);
    *dest = Py_None;
  }
  return FALSE;
}

SIDL_string__python_clone_array_RETURN
SIDL_string__python_clone_array SIDL_string__python_clone_array_PROTO
{
  PyObject *pya = NULL;
  if (array) {
    const int dimen = SIDL_string__array_dimen(array);
    int i;
    int *lower = malloc(sizeof(int) * dimen);
    int *upper = malloc(sizeof(int) * dimen);
    int *numelem = malloc(sizeof(int) * dimen);
    for(i = 0; i < dimen; ++i) {
      lower[i] = SIDL_string__array_lower(array, i);
      upper[i] = SIDL_string__array_upper(array, i);
      numelem[i] = 1 + upper[i] - lower[i];
    }
    pya = PyArray_FromDims(dimen, numelem, PyArray_OBJECT);
    if (pya) {
      if (!SIDL_array__convert_sidl(pya, dimen, lower, upper, 
                                    numelem, array, getAndConvertString)) {
        Py_DECREF(pya);
        pya = NULL;
      }
    }
    free(numelem);
    free(upper);
    free(lower);
  }
  return pya;
}

static int
getAndConvertOpaque(void *array, const int *ind, PyObject **dest)
{
  void *vptr =
    SIDL_opaque__array_get((struct SIDL_opaque__array *)array, ind);
  *dest = PyCObject_FromVoidPtr(vptr, NULL);
  return FALSE;
}

SIDL_opaque__python_clone_array_RETURN
SIDL_opaque__python_clone_array SIDL_opaque__python_clone_array_PROTO
{
  PyObject *pya = NULL;
  if (array) {
    const int dimen = SIDL_opaque__array_dimen(array);
    int i;
    int *lower = malloc(sizeof(int) * dimen);
    int *upper = malloc(sizeof(int) * dimen);
    int *numelem = malloc(sizeof(int) * dimen);
    for(i = 0; i < dimen; ++i) {
      lower[i] = SIDL_opaque__array_lower(array, i);
      upper[i] = SIDL_opaque__array_upper(array, i);
      numelem[i] = 1 + upper[i] - lower[i];
    }
    pya = PyArray_FromDims(dimen, numelem, PyArray_OBJECT);
    if (pya) {
      if (!SIDL_array__convert_sidl(pya, dimen, lower, upper,
                                    numelem, array, getAndConvertOpaque)) {
        Py_DECREF(pya);
        pya = NULL;
      }
    }
    free(numelem);
    free(upper);
    free(lower);
  }
  return pya;
}

SIDL_bool__python_borrow_array_RETURN
SIDL_bool__python_borrow_array SIDL_bool__python_borrow_array_PROTO
{ return SIDL_bool__python_clone_array(array); }

SIDL_char__python_borrow_array_RETURN
SIDL_char__python_borrow_array SIDL_char__python_borrow_array_PROTO
{ return SIDL_char__python_clone_array(array); }

SIDL_dcomplex__python_borrow_array_RETURN
SIDL_dcomplex__python_borrow_array SIDL_dcomplex__python_borrow_array_PROTO
{ return SIDL_dcomplex__python_clone_array(array); }

SIDL_double__python_borrow_array_RETURN
SIDL_double__python_borrow_array SIDL_double__python_borrow_array_PROTO
{ return SIDL_double__python_clone_array(array); }

SIDL_fcomplex__python_borrow_array_RETURN
SIDL_fcomplex__python_borrow_array SIDL_fcomplex__python_borrow_array_PROTO
{ return SIDL_fcomplex__python_clone_array(array); }

SIDL_float__python_borrow_array_RETURN
SIDL_float__python_borrow_array SIDL_float__python_borrow_array_PROTO
{ return SIDL_float__python_clone_array(array); }

SIDL_int__python_borrow_array_RETURN
SIDL_int__python_borrow_array SIDL_int__python_borrow_array_PROTO
{ return SIDL_int__python_clone_array(array); }

SIDL_long__python_borrow_array_RETURN
SIDL_long__python_borrow_array SIDL_long__python_borrow_array_PROTO
{ return SIDL_long__python_clone_array(array); }

SIDL_opaque__python_borrow_array_RETURN
SIDL_opaque__python_borrow_array SIDL_opaque__python_borrow_array_PROTO
{ return SIDL_opaque__python_clone_array(array); }

SIDL_string__python_borrow_array_RETURN
SIDL_string__python_borrow_array SIDL_string__python_borrow_array_PROTO
{ return SIDL_string__python_clone_array(array); }

#ifdef __cplusplus
extern "C" void initSIDLPyArrays(void);
#else
extern void initSIDLPyArrays(void);
#endif

static struct PyMethodDef spa_methods[] = {
  /* this module exports no methods */
  { NULL, NULL }
};

void
initSIDLPyArrays(void) 
{
  PyObject *module, *dict, *c_api;
  static void *spa_api[SIDLPyArrays_API_pointers];
  module = Py_InitModule((char *) "SIDLPyArrays", spa_methods);
  import_array();
  dict = PyModule_GetDict(module);
  spa_api[SIDL_bool__borrow_python_array_NUM] =
    (void *)SIDL_bool__borrow_python_array;
  spa_api[SIDL_bool__clone_python_array_NUM] =
    (void *)SIDL_bool__clone_python_array;
  spa_api[SIDL_bool__python_borrow_array_NUM] =
    (void *)SIDL_bool__python_borrow_array;
  spa_api[SIDL_bool__python_clone_array_NUM] =
    (void *)SIDL_bool__python_clone_array;
  spa_api[SIDL_bool__python_destroy_array_NUM] =
    (void *)SIDL_bool__array_destroy;
  spa_api[SIDL_char__borrow_python_array_NUM] =
    (void *)SIDL_char__borrow_python_array;
  spa_api[SIDL_char__clone_python_array_NUM] =
    (void *)SIDL_char__clone_python_array;
  spa_api[SIDL_char__python_borrow_array_NUM] =
    (void *)SIDL_char__python_borrow_array;
  spa_api[SIDL_char__python_clone_array_NUM] =
    (void *)SIDL_char__python_clone_array;
  spa_api[SIDL_char__python_destroy_array_NUM] =
    (void *)SIDL_char__array_destroy;
  spa_api[SIDL_dcomplex__borrow_python_array_NUM] =
    (void *)SIDL_dcomplex__borrow_python_array;
  spa_api[SIDL_dcomplex__clone_python_array_NUM] =
    (void *)SIDL_dcomplex__clone_python_array;
  spa_api[SIDL_dcomplex__python_borrow_array_NUM] =
    (void *)SIDL_dcomplex__python_borrow_array;
  spa_api[SIDL_dcomplex__python_clone_array_NUM] =
    (void *)SIDL_dcomplex__python_clone_array;
  spa_api[SIDL_dcomplex__python_destroy_array_NUM] =
    (void *)SIDL_dcomplex__array_destroy;
  spa_api[SIDL_double__borrow_python_array_NUM] =
    (void *)SIDL_double__borrow_python_array;
  spa_api[SIDL_double__clone_python_array_NUM] =
    (void *)SIDL_double__clone_python_array;
  spa_api[SIDL_double__python_borrow_array_NUM] =
    (void *)SIDL_double__python_borrow_array;
  spa_api[SIDL_double__python_clone_array_NUM] =
    (void *)SIDL_double__python_clone_array;
  spa_api[SIDL_double__python_destroy_array_NUM] =
    (void *)SIDL_double__array_destroy;
  spa_api[SIDL_fcomplex__borrow_python_array_NUM] =
    (void *)SIDL_fcomplex__borrow_python_array;
  spa_api[SIDL_fcomplex__clone_python_array_NUM] =
    (void *)SIDL_fcomplex__clone_python_array;
  spa_api[SIDL_fcomplex__python_borrow_array_NUM] =
    (void *)SIDL_fcomplex__python_borrow_array;
  spa_api[SIDL_fcomplex__python_clone_array_NUM] =
    (void *)SIDL_fcomplex__python_clone_array;
  spa_api[SIDL_fcomplex__python_destroy_array_NUM] =
    (void *)SIDL_fcomplex__array_destroy;
  spa_api[SIDL_float__borrow_python_array_NUM] =
    (void *)SIDL_float__borrow_python_array;
  spa_api[SIDL_float__clone_python_array_NUM] =
    (void *)SIDL_float__clone_python_array;
  spa_api[SIDL_float__python_borrow_array_NUM] =
    (void *)SIDL_float__python_borrow_array;
  spa_api[SIDL_float__python_clone_array_NUM] =
    (void *)SIDL_float__python_clone_array;
  spa_api[SIDL_float__python_destroy_array_NUM] =
    (void *)SIDL_float__array_destroy;
  spa_api[SIDL_int__borrow_python_array_NUM] =
    (void *)SIDL_int__borrow_python_array;
  spa_api[SIDL_int__clone_python_array_NUM] =
    (void *)SIDL_int__clone_python_array;
  spa_api[SIDL_int__python_borrow_array_NUM] =
    (void *)SIDL_int__python_borrow_array;
  spa_api[SIDL_int__python_clone_array_NUM] =
    (void *)SIDL_int__python_clone_array;
  spa_api[SIDL_int__python_destroy_array_NUM] =
    (void *)SIDL_int__array_destroy;
  spa_api[SIDL_long__borrow_python_array_NUM] =
    (void *)SIDL_long__borrow_python_array;
  spa_api[SIDL_long__clone_python_array_NUM] =
    (void *)SIDL_long__clone_python_array;
  spa_api[SIDL_long__python_borrow_array_NUM] =
    (void *)SIDL_long__python_borrow_array;
  spa_api[SIDL_long__python_clone_array_NUM] =
    (void *)SIDL_long__python_clone_array;
  spa_api[SIDL_long__python_destroy_array_NUM] =
    (void *)SIDL_long__array_destroy;
  spa_api[SIDL_opaque__borrow_python_array_NUM] =
    (void *)SIDL_opaque__borrow_python_array;
  spa_api[SIDL_opaque__clone_python_array_NUM] =
    (void *)SIDL_opaque__clone_python_array;
  spa_api[SIDL_opaque__python_borrow_array_NUM] =
    (void *)SIDL_opaque__python_borrow_array;
  spa_api[SIDL_opaque__python_clone_array_NUM] =
    (void *)SIDL_opaque__python_clone_array;
  spa_api[SIDL_opaque__python_destroy_array_NUM] =
    (void *)SIDL_opaque__array_destroy;
  spa_api[SIDL_string__borrow_python_array_NUM] =
    (void *)SIDL_string__borrow_python_array;
  spa_api[SIDL_string__clone_python_array_NUM] =
    (void *)SIDL_string__clone_python_array;
  spa_api[SIDL_string__python_borrow_array_NUM] =
    (void *)SIDL_string__python_borrow_array;
  spa_api[SIDL_string__python_clone_array_NUM] =
    (void *)SIDL_string__python_clone_array;
  spa_api[SIDL_string__python_destroy_array_NUM] =
    (void *)SIDL_string__array_destroy;
  spa_api[SIDL_array__convert_python_NUM] =
    (void *)SIDL_array__convert_python;
  spa_api[SIDL_array__convert_sidl_NUM] =
    (void *)SIDL_array__convert_sidl;
  spa_api[SIDL_array__extract_python_info_NUM] =
    (void *)SIDL_array__extract_python_info;
  c_api = PyCObject_FromVoidPtr((void *)spa_api, NULL);
  if (c_api) {
    PyDict_SetItemString(dict, (char *) "_C_API", c_api);
    Py_DECREF(c_api);
  }
  if (PyErr_Occurred()) {
    Py_FatalError((char *) "Can't initialize module SIDLPyArrays.");
  }
}
